﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;


/// <summary>
/// Gコード中の高さ・移動量(合計)・Eの値(合計)を求める
/// これにより、ある高さでの吐出量を計算する
/// Zの値は整数、移動量とEはdoubleで保持する
/// </summary>
namespace CalcEValue {
	class ParseFile {
		Double prevHeight_ = 0.0;
		Double prevX_ = 0.0;	// 座標値
		Double prevY_ = 0.0;	// 座標値
		Double prevE_ = 0.0;	// 吐出量
		Double headLength_ = 0.0;  // ヘッドの移動距離
		Double filamentLength_ = 0.0;   // フィラメントの移動距離
		Double headHeight_ = 0.0;
		bool isAbslute_ = true; // 絶対座標モードかどうか

		StreamWriter csvFile_;

		public ParseFile(StreamWriter csvFile) {
			csvFile_ = csvFile;
		}
		/// <summary>
		/// 文字列の中から英文字(1文字)をコマンドとして、その後に続く数値と.と-を値として解析して返す
		/// </summary>
		/// <param name="line"></param>
		/// <param name="command">コマンド文字を小文字にしたもの</param>
		/// <param name="value">数値らしき値</param>
		/// <returns>commandとvalueを除いた値</returns>
		string parseCommand(string line, out char command, out string value) {
			var work = line.Trim();
			command = ' ';
			value = "";
			if (work.Length > 0) {
				command = work[0];
				int next = 1;
				for (; next < work.Length && (Char.IsDigit(work[next]) || work[next] == '-' || work[next] == '.'); next++) {
						value += work[next];
				}
				work = work.Substring(next, work.Length - next);
			}
			return work;
		}

		static Double distance(Double startx, Double starty, Double endx, Double endy) {
			var deltaX = startx - endx;
			var deltaY = starty - endy;
			return Math.Sqrt(deltaX * deltaX + deltaY * deltaY);
		}

		/// <summary>
		/// XとYの値が一致しているかどうかを判断する。
		/// 根拠とするε値は、小数点4桁まで判断可能な値にする
		/// </summary>
		/// <param name="x"></param>
		/// <param name="y"></param>
		/// <returns></returns>
		static bool isEqual(Double x, Double y) {
			return Math.Abs(x - y) < 0.0001;
		}

		/// <summary>
		/// 文字列を実数に変換し、前回値の更新を行う
		/// 一応、内部的に保持している絶対値・相対値情報を使う
		/// </summary>
		/// <param name="prev">前の値</param>
		/// <param name="value">今回の値</param>
		/// <returns>計算結果</returns>
		Double moveValue(Double prev, string value) {
			Double work;
			if (Double.TryParse(value, out work)) {
				if (isAbslute_) {
					return work;
				}
				else {
					return prev + work;
				}
			}
			return prev;
		}

		public void post() {
			if (headLength_ > 0.0) {
				csvFile_.WriteLine("{0},{1},{2},{3}", headHeight_, headLength_, filamentLength_, filamentLength_ * 0.875 * 0.875 * Math.PI / headLength_);
			}
		}

		public void pre() {
			csvFile_.WriteLine("高さ,描画長さ,素材長さ,1mm描画時の素材体積");
		}

		public void lineParse(string line) {
			Double currentX = prevX_;
			Double currentY = prevY_;
			Double currentHeight = prevHeight_;
			Double currentE = prevE_;
			long gMode = 0;
			bool isExtruede = false;

			// 小文字化してから空白と;以降をコメントとして除去する
			var work = line.Trim().ToLower();
			if (work.IndexOf(';') >= 0) {
				work = work.Remove(work.IndexOf(';'));
			}
			// 1行の解釈
			while (work.Length > 0) {
				char command;
				string value;
				work = parseCommand(work, out command, out value);
				if (command == ' ') {
					// データが読み込まれていない場合なので
					break;
				}
				switch (command) {
				case 'g':
					switch (value) {
					case "1":
						// ヘッド移動モード
						gMode = 1;
						break;
					case "92":
						gMode = 92;
						// 原点設定(主にEの値)
						break;
					}
					break;
				case 'f':
					// 速度指示は何もしない
					break;
				case 'x':
					currentX = moveValue(prevX_, value);
					break;
				case 'y':
					currentY = moveValue(prevY_, value);
					break;
				case 'z':
					currentHeight = moveValue(prevHeight_, value);
					break;
				case 'e':
					if (gMode == 92) {
						prevE_ = moveValue(0.0, value);
					}
					else {
						currentE = moveValue(prevE_, value);
						if (prevE_ < currentE) {
							// 吐出量が正の場合のみ、計算対象とする
							isExtruede = true;
						}
					}
					break;
				case 'm':
					switch(value) {
					case "91":
						// 相対移動
						isAbslute_ = false;
						break;
					case "90":
						// 絶対移動
						isAbslute_ = true;
						break;
					}
					break;
				}
			}
			if (gMode == 1 && isExtruede) {
				if (!isEqual(headHeight_, currentHeight)) {
					//　高さが異なる場合、ここでいったん切る
					post();
					headLength_ = 0.0;
					filamentLength_ = 0.0;
					headHeight_ = currentHeight;
				}
				// 移動の処理
				headLength_ += distance(prevX_, prevY_, currentX, currentY);
				filamentLength_ += currentE - prevE_;
				prevE_ = currentE;
			}
			prevX_ = currentX;
			prevY_ = currentY;
			prevHeight_ = currentHeight;
		}
	}
	class CalcEValue {

		static void Main(string[] args) {
			foreach (var fname in args) {
				using (StreamReader fp = new StreamReader(fname)) {
					// ファイルのパース処理
					using (StreamWriter csvFile = new StreamWriter(System.IO.Path.ChangeExtension(fname, "csv"), false, Encoding.GetEncoding("shift_jis"))) {
						ParseFile parser = new ParseFile(csvFile);
						parser.pre();
						string line;
						while ((line = fp.ReadLine()) != null) {
							parser.lineParse(line);
						}
						parser.post();

					}
				}
			}
		}

	}
}
